unit Main;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, IdWhoIsServer, DaemonApp, eventlog,
  mysql50, {$IFDEF WINDOWS} registry {$ELSE} IniFiles {$ENDIF},
  IdSocketHandle, IdCustomTCPServer, IdContext, IdComponent, LCLProc;

type

  { Twhoisd }

  Twhoisd = class(TDaemon)
    EventLog: TEventLog;
    IdWhoIsServer: TIdWhoIsServer;
    procedure DataModuleExecute(Sender: TCustomDaemon);
    procedure DataModuleStart(Sender: TCustomDaemon; var OK: Boolean);
    procedure DataModuleStop(Sender: TCustomDaemon; var OK: Boolean);
    procedure EventLogGetCustomCategory(Sender: TObject; var Code: Word);
    procedure IdWhoIsServerAfterBind(Sender: TObject);
    procedure IdWhoIsServerBeforeBind(AHandle: TIdSocketHandle);
    procedure IdWhoIsServerConnect(AContext: TIdContext);
    procedure IdWhoIsServerException(AContext: TIdContext; AException: Exception
      );
    procedure IdWhoIsServerStatus(ASender: TObject; const AStatus: TIdStatus;
      const AStatusText: string);
  private
    fPort:word;
    fLogLevel:byte;
    fUseMySQL:boolean;
    fMySQLDB: String;
    fMySQLScheme: String;
    fMySQLUser: String;
    fMySQLPassword: String;
    fMySQL:PMYSQL;
    function GetParameters:boolean;
    function LoadConfig: boolean;
    procedure CommandLookup(AContext:TIdContext; ALookup: string; var VResponse : String);
    { private declarations }
  public
    { public declarations }
  end; 

var
  whoisd: Twhoisd;

const
  sWhoisdConf: string = '/etc/whoisd.conf';

implementation

procedure RegisterDaemon; 
begin
  RegisterDaemonClass(Twhoisd)
end;

function StrReplace(aSrc, aFrom, aTo: string):string;
var
  i:integer;
begin
  i:=Pos(aFrom,aSrc);
  while i<>0 do
  begin
    Delete(aSrc,i,length(aFrom));
    Insert(aSrc,aTo,i);
    i:=Pos(aFrom,aSrc);
  end;
  result:=aSrc;
end;

{ Twhoisd }

procedure Twhoisd.DataModuleExecute(Sender: TCustomDaemon);
begin

end;

procedure Twhoisd.DataModuleStart(Sender: TCustomDaemon; var OK: Boolean);
begin
  If GetParameters then
  begin
    EventLog.Active:=true;
    if(fLogLevel>=1) then EventLog.Info('Start');
    if LoadConfig then
    begin
      if fUseMySQL then
      begin
        fMySQL:=mysql_init(nil);
        fMySQL:=mysql_real_connect(fMySQL,PChar(fMySQLDB),PChar(fMySQLUser),PChar(fMySQLPassword),PChar(fMySQLScheme),0,nil,0);
        if fMySQL=nil then
        begin
          if(fLogLevel>=2) then EventLog.Warning('Can''t connect to database %s as user %s',[fMySQLDB,fMySQLUser]);
          Stop;
        end;
        mysql_set_character_set(fMySQL,PChar('utf8'));
      end;
      try
        IdWhoIsServer.OnCommandLookup:=@CommandLookup;
        IdWhoIsServer.Active:=true;
      except
        if(fLogLevel>=3) then EventLog.Error('Fail to bind socket');
        Stop;
      end;
    end else Stop;
  end else Stop;
end;

procedure Twhoisd.DataModuleStop(Sender: TCustomDaemon; var OK: Boolean);
begin
  if(fLogLevel>=1) then EventLog.Info('Stop');
  IdWhoIsServer.Active:=false;
  EventLog.Active:=false;
  Application.Terminate;
end;

procedure Twhoisd.EventLogGetCustomCategory(Sender: TObject; var Code: Word);
begin

end;

procedure Twhoisd.IdWhoIsServerAfterBind(Sender: TObject);
begin
  if(fLogLevel>=1) then EventLog.Info('Start listen');
end;

procedure Twhoisd.IdWhoIsServerBeforeBind(AHandle: TIdSocketHandle);
var
  i:integer;
begin
  if(fLogLevel>=1) then EventLog.Info('Bind interfaces:');
  for i:=0 to IdWhoIsServer.Bindings.Count-1 do
    if(fLogLevel>=1) then EventLog.Info(IdWhoIsServer.Bindings[i].IP+':'+IntToStr(IdWhoIsServer.Bindings[i].Port));
end;

procedure Twhoisd.IdWhoIsServerConnect(AContext: TIdContext);
var
  sRequestURL:string;
begin
  if(fLogLevel>=0) then EventLog.Debug('Client connected %s:%d',[AContext.Binding.PeerIP,AContext.Binding.Port]);
end;

procedure Twhoisd.IdWhoIsServerException(AContext: TIdContext;
  AException: Exception);
begin
  EventLog.Error(AException.Message);
end;

procedure Twhoisd.IdWhoIsServerStatus(ASender: TObject;
  const AStatus: TIdStatus; const AStatusText: string);
begin

end;

function Twhoisd.GetParameters:boolean;
var
  i:integer;
  aval:Integer;
begin
  fPort:=43;
  fLogLevel:=2;
  result:=true;
  i:=1;
  while i<Paramcount do
  begin
    if ParamStr(i)='-p' then
    begin
      if (Paramcount<i+1) or not TryStrToInt(Paramstr(i+1),aval) then
      begin
        result:=false;
        break;
      end;
      fPort:=aval;
      inc(i);
    end else
    if ParamStr(i)='-c' then
    begin
      if (Paramcount<i+1) then
      begin
        result:=false;
        break;
      end;
      sWhoisdConf:=Paramstr(i+1);
      inc(i);
    end else
    if ParamStr(i)='-l' then
    begin
      aval:=0;
      if (Paramcount<i+1) or (not TryStrToInt(Paramstr(i+1),aval)) or (aval<0) or (aval>3) then
      begin
        result:=false;
        break;
      end;
      fLogLevel:=aval;
      inc(i);
    end else
    if ParamStr(i)='-h' then
    begin
      writeln('Usage: '+ExtractFileName(ParamStr(0))+' [options] <command>');
      writeln('  Commands:');
      writeln('    install   - installs daemon');
      writeln('    run       - run daemon instance');
      writeln('    uninstall - remove daemon');
      writeln();
      writeln('  Options:');
      writeln('    -h        - Show this help');
      writeln('    -l <lev>  - Set log level from 0 to 3, default 2');
      writeln('    -c <file> - Set config file, default is /etc/wh');
      writeln('    -p <port> - Set daemon listening port, default 43');
      result:=false;
      break;
    end;
    inc(i);
  end;
end;

function Twhoisd.LoadConfig: boolean;
var
  s,SectName:string;
  aPort:word;
  sl:TStringList;
  i: Integer;
begin
  result:=false;
  {$IFDEF WINDOWS}
  SectName:='Software\Whoisd';
  with TRegIniFile.Create(HKEY_LOCAL_MACHINE) do
  {$ELSE}
  SectName:='Daemon';
  if FileExists(sWhoisdConf) and FileIsReadable(sWhoisdConf) then
    with TIniFile.Create(sWhoisdConf) do
  {$ENDIF}
    begin
      IdWhoIsServer.Bindings.Clear;
      aPort:=ReadInteger(SectName,'Port',fPort);
      s:=ReadString(SectName,'Bind','');
      s:=StrReplace(s,' ','');
      if s='' then
      begin
        with IdWhoIsServer.Bindings.Add do
        begin
          IP:='0.0.0.0';
          Port:=aPort;
        end;
      end else
      begin
        sl:=TStringList.Create;
        sl.Delimiter:=',';
        sl.DelimitedText:=s;
        for i:=0 to sl.Count-1 do
        with IdWhoIsServer.Bindings.Add do
        begin
          IP:=sl[i];
          Port:=aPort;
        end;
        sl.Free;
      end;
      fUseMySQL:=ReadBool(SectName,'UseMySQL',false);
      fMySQLDB:=ReadString(SectName,'MySQLDB','localhost');
      fMySQLScheme:=ReadString(SectName,'MySQLScheme','whois');
      fMySQLUser:=ReadString(SectName,'MySQLUser','whois');
      fMySQLPassword:=ReadString(SectName,'MySQLPwd','');
      result:=true;
      Free;
    end;
end;

procedure Twhoisd.CommandLookup(AContext: TIdContext; ALookup: string;
  var VResponse: String);
var
  aHasDomain: boolean;
  s,SectName:string;
  aPort:word;
  sl:TStringList;
  i: Integer;
  aMySQL:PMYSQL;
  recbuf : PMYSQL_RES;
  rowbuf : MYSQL_ROW;
begin
  ALookup:=StrReplace(UTF8LowerCase(ALookup),'"','''');
  if(fLogLevel>=0) then EventLog.Debug('Requested domain %s',[ALookup]);
  aHasDomain:=false;
  sl:=TStringList.Create;
  if fUseMySQL then
  begin
    aMySQL:=nil;
    if mysql_ping(fMySQL)<>0 then
    begin
      aMySQL:=mysql_real_connect(fMySQL,PChar(fMySQLDB),PChar(fMySQLUser),PChar(fMySQLPassword),PChar(fMySQLScheme),0,nil,0);
      if aMySQL<>nil then
      begin
        fMySQL:=aMySQL;
        mysql_set_character_set(fMySQL,PChar('utf8'));
      end else
        if(fLogLevel>=2) then EventLog.Warning('Can''t connect to database %s as user %s',[fMySQLDB,fMySQLUser]);
    end else aMySQL:=fMySQL;
    if aMySQL<>nil then
      if mysql_query(aMySQL,PChar('SELECT `domain_name`, `created`, `updated`, `expire`, `rname`, `rgender`, `rorg`, `rstreet1`, `rstreet2`, `rcity`, `rstate`, `rzip`, `rcountry`, `rphone`, `rphoneext`, `rfax`, `rfaxext`, `remail`, '+
         '`aname`, `agender`, `aorg`, `astreet1`, `astreet2`, `acity`, `astate`, `azip`, `acountry`, `aphone`, `aphoneext`, `afax`, `afaxext`, `aemail`, '+
         '`tname`, `tgender`, `torg`, `tstreet1`, `tstreet2`, `tcity`, `tstate`, `tzip`, `tcountry`, `tphone`, `tphoneext`, `tfax`, `tfaxext`, `temail` '+
         ' FROM whois WHERE `domain_name`="'+ALookup+'"'))>=0 then
      begin
        recbuf:=mysql_store_result(aMySQL);
        if recbuf<>nil then
        begin
          rowbuf:=mysql_fetch_row(recbuf);
          if rowbuf<>nil then
          begin
            aHasDomain:=true;
            sl.Add('Domain Name: '+rowbuf[0]);
            sl.Add('Created On: '+rowbuf[1]);
            sl.Add('Last Updated On: '+rowbuf[2]);
            sl.Add('Expiration Date: '+rowbuf[3]);
            sl.Add('Registrant Name: '+rowbuf[4]);
            sl.Add('Registrant Gender: '+rowbuf[5]);
            sl.Add('Registrant Organisation: '+rowbuf[6]);
            sl.Add('Registrant Street1: '+rowbuf[7]);
            sl.Add('Registrant Street2: '+rowbuf[8]);
            sl.Add('Registrant City: '+rowbuf[9]);
            sl.Add('Registrant State/Province: '+rowbuf[10]);
            sl.Add('Registrant Postal Code: '+rowbuf[11]);
            sl.Add('Registrant Country: '+rowbuf[12]);
            sl.Add('Registrant Phone: '+rowbuf[13]);
            sl.Add('Registrant Phone Ext.: '+rowbuf[14]);
            sl.Add('Registrant FAX: '+rowbuf[15]);
            sl.Add('Registrant FAX Ext.: '+rowbuf[16]);
            sl.Add('Registrant Email: '+rowbuf[17]);
            sl.Add('Admin Name: '+rowbuf[18]);
            sl.Add('Admin Gender: '+rowbuf[19]);
            sl.Add('Admin Organisation: '+rowbuf[20]);
            sl.Add('Admin Street1: '+rowbuf[21]);
            sl.Add('Admin Street2: '+rowbuf[22]);
            sl.Add('Admin City: '+rowbuf[23]);
            sl.Add('Admin State/Province: '+rowbuf[24]);
            sl.Add('Admin Postal Code: '+rowbuf[25]);
            sl.Add('Admin Country: '+rowbuf[26]);
            sl.Add('Admin Phone: '+rowbuf[27]);
            sl.Add('Admin Phone Ext.: '+rowbuf[28]);
            sl.Add('Admin FAX: '+rowbuf[29]);
            sl.Add('Admin FAX Ext.: '+rowbuf[30]);
            sl.Add('Admin Email: '+rowbuf[31]);
            sl.Add('Tech Name: '+rowbuf[32]);
            sl.Add('Tech Gender: '+rowbuf[33]);
            sl.Add('Tech Organisation: '+rowbuf[34]);
            sl.Add('Tech Street1: '+rowbuf[35]);
            sl.Add('Tech Street2: '+rowbuf[36]);
            sl.Add('Tech City: '+rowbuf[37]);
            sl.Add('Tech State/Province: '+rowbuf[38]);
            sl.Add('Tech Postal Code: '+rowbuf[39]);
            sl.Add('Tech Country: '+rowbuf[40]);
            sl.Add('Tech Phone: '+rowbuf[41]);
            sl.Add('Tech Phone Ext.: '+rowbuf[42]);
            sl.Add('Tech FAX: '+rowbuf[43]);
            sl.Add('Tech FAX Ext.: '+rowbuf[44]);
            sl.Add('Tech Email: '+rowbuf[45]);
          end;
          mysql_free_result(recbuf);
        end;
      end else
        if(fLogLevel>=2) then EventLog.Warning('Can''t execute SELECT query form `whois` table');
  end;
  if not aHasDomain then
  begin
    {$IFDEF WINDOWS}
    SectName:='Software\Whoisd\'+ALookup;
    with TRegIniFile.Create(HKEY_LOCAL_MACHINE) do
    {$ELSE}
    SectName:=ALookup;
    if FileExists(sWhoisdConf) and FileIsReadable(sWhoisdConf) then
      with TIniFile.Create(sWhoisdConf) do
    {$ENDIF}
    begin
      if SectionExists(SectName) then
      begin
        aHasDomain:=true;
        sl.Add('Domain Name: '+SectName);
        sl.Add('Created On: '+ReadString(SectName,'Created',''));
        sl.Add('Last Updated On: '+ReadString(SectName,'Updated',''));
        sl.Add('Expiration Date: '+ReadString(SectName,'Expire',''));
        sl.Add('Registrant Name: '+ReadString(SectName,'rName',''));
        sl.Add('Registrant Gender: '+ReadString(SectName,'rGender',''));
        sl.Add('Registrant Organisation: '+ReadString(SectName,'rOrg',''));
        sl.Add('Registrant Street1: '+ReadString(SectName,'rStreet1',''));
        sl.Add('Registrant Street2: '+ReadString(SectName,'rStreet2',''));
        sl.Add('Registrant City: '+ReadString(SectName,'rCity',''));
        sl.Add('Registrant State/Province: '+ReadString(SectName,'rState',''));
        sl.Add('Registrant Postal Code: '+ReadString(SectName,'rZip',''));
        sl.Add('Registrant Country: '+ReadString(SectName,'rCountry',''));
        sl.Add('Registrant Phone: '+ReadString(SectName,'rPhone',''));
        sl.Add('Registrant Phone Ext.: '+ReadString(SectName,'rPhoneExt',''));
        sl.Add('Registrant FAX: '+ReadString(SectName,'rFax',''));
        sl.Add('Registrant FAX Ext.: '+ReadString(SectName,'rFaxExt',''));
        sl.Add('Registrant Email: '+ReadString(SectName,'rEmail',''));
        sl.Add('Admin Name: '+ReadString(SectName,'aName',''));
        sl.Add('Admin Gender: '+ReadString(SectName,'aGender',''));
        sl.Add('Admin Organisation: '+ReadString(SectName,'aOrg',''));
        sl.Add('Admin Street1: '+ReadString(SectName,'aStreet1',''));
        sl.Add('Admin Street2: '+ReadString(SectName,'aStreet2',''));
        sl.Add('Admin City: '+ReadString(SectName,'arCity',''));
        sl.Add('Admin State/Province: '+ReadString(SectName,'aState',''));
        sl.Add('Admin Postal Code: '+ReadString(SectName,'aZip',''));
        sl.Add('Admin Country: '+ReadString(SectName,'aCountry',''));
        sl.Add('Admin Phone: '+ReadString(SectName,'aPhone',''));
        sl.Add('Admin Phone Ext.: '+ReadString(SectName,'aPhoneExt',''));
        sl.Add('Admin FAX: '+ReadString(SectName,'aFax',''));
        sl.Add('Admin FAX Ext.: '+ReadString(SectName,'aFaxExt',''));
        sl.Add('Admin Email: '+ReadString(SectName,'aEmail',''));
        sl.Add('Tech Name: '+ReadString(SectName,'tName',''));
        sl.Add('Tech Gender: '+ReadString(SectName,'tGender',''));
        sl.Add('Tech Organisation: '+ReadString(SectName,'tOrg',''));
        sl.Add('Tech Street1: '+ReadString(SectName,'tStreet1',''));
        sl.Add('Tech Street2: '+ReadString(SectName,'tStreet2',''));
        sl.Add('Tech City: '+ReadString(SectName,'tCity',''));
        sl.Add('Tech State/Province: '+ReadString(SectName,'tState',''));
        sl.Add('Tech Postal Code: '+ReadString(SectName,'tZip',''));
        sl.Add('Tech Country: '+ReadString(SectName,'tCountry',''));
        sl.Add('Tech Phone: '+ReadString(SectName,'tPhone',''));
        sl.Add('Tech Phone Ext.: '+ReadString(SectName,'tPhoneExt',''));
        sl.Add('Tech FAX: '+ReadString(SectName,'tFax',''));
        sl.Add('Tech FAX Ext.: '+ReadString(SectName,'tFaxExt',''));
        sl.Add('Tech Email: '+ReadString(SectName,'tEmail',''));
        sl.Add('Name Server: '+ReadString(SectName,'ns1',''));
        sl.Add('Name Server: '+ReadString(SectName,'ns2',''));
        sl.Add('Name Server: '+ReadString(SectName,'ns3',''));
        sl.Add('Name Server: '+ReadString(SectName,'ns4',''));
        sl.Add('Name Server: '+ReadString(SectName,'ns5',''));
        sl.Add('Name Server: '+ReadString(SectName,'ns6',''));
        sl.Add('Name Server: '+ReadString(SectName,'ns7',''));
        sl.Add('Name Server: : '+ReadString(SectName,'ns8',''));
        sl.Add('Name Server: '+ReadString(SectName,'ns9',''));
        sl.Add('Name Server: '+ReadString(SectName,'ns10',''));
      end;
      Free;
    end;
  end;
  if aHasDomain then
    VResponse:=sl.Text
  else
    VResponse:='NOT FOUND';
  sl.Free;
end;

{$R *.lfm}


initialization
  RegisterDaemon; 
end.

